import numpy as np


# Basic Functions
def linspace(starts, stops, nums):
    """
    2/23/22 VERIFIED

    A modified version of NumPy's linspace. Input start and endpoints can be single integers or a list (though not a
    mismatch). The outputs can be collected as a, b, c, ... = linspace(...) if lists are given as input. If integers
    are given as input, np.linspace is called directly.

    :param starts: int or list
        Starting points of the linspace arrays
    :param stops: int or list
        End points of the linspace arrays
    :param nums: int or list
        Number of steps in each array

    :return: output : ndarray
        The array(s) evaluated generated from the modified linspace

    --------
    Examples
    --------
    a, b = linspace([1, 3], [2, 4], [8, 5]) <--- two arrays with 8 and 5 entries, starting at 1 and 3, ending at 2
                                                    and 4 respectively
    a, b = linspace([1, 3], [2, 4], [8]) <--- two arrays with 8 entries, starting at 1 and 3, ending at 2 and 4
                                                respectively
    a = linspace([1], [2], [8]) <--- For consistency to above cases (though a standard numpy call is possible)
    a = linspace(1, 2, 8) <--- Looks like the standard numpy call!
    """
    if isinstance(starts, list) and isinstance(stops, list) and isinstance(nums, list):
        # Getting the lengths of the inputs that at this point are all lists.
        min_len, max_len, num_len = min([len(starts), len(stops)]), max([len(starts), len(stops)]), len(nums)

        # Finding the amount of unique lengths. 3 doesn't make sense, so it is a failure case.
        if len({min_len, max_len, num_len}) == 3:
            print('Unable to interpret ez.linspace inputs (3 lists of different sizes where given')
            return None
        # If the amount of unqiue lengths is 1, then the default numpy call is returned
        elif len({min_len, max_len, num_len}) == 1:
            return np.linspace(starts[0], stops[0], num=nums[0])
        # If the amount of unique lengths is 2, then the modified numpy call is returned
        else:
            # Converting all the lists to np arrays of the same size
            ones_len = max([max_len, num_len])
            starts = np.ones(ones_len) * starts
            stops = np.ones(ones_len) * stops
            nums = np.ones(ones_len) * nums
            steps = (stops - starts) / (nums - 1)
            a = list()
            # Collecting all the linspace calls, we use np.arange here because why not
            for i in range(ones_len):
                a.append(np.arange(nums[i]) * steps[i] + starts[i])
            return a
    else:
        # If everything passed was an integer, just do the default numpy linspace call
        return np.linspace(starts, stops, num=nums)


def meshgrid(array_1d, dim=2, indexing='xy'):
    """
    2/23/22 VERIFIED

    A modified version of meshgrid that only requires 1 input array. Often, we are just wanting square grids,
    so this is just a shortcut where the input array_1d is a single linspace that can be upped to a meshgrid via the
    dim value.

    :param array_1d: ndarray or list
        The array you want to replicated multiple times to form a meshgrid
    :param dim: int
        The number of dimensions you want a meshgrid over
    :param indexing:
        The indexing. Cartesian is default and preferred
    :return: output: default 'xy' or optionally 'ij'
        The output meshgrid

    --------
    Examples
    --------
    a, b, c = meshgrid([-10, -5, 0, 5, 10], 3)
    aa, bb = meshgrid(ez.linspace(-10, 10, 30), 2)
    """
    # It's that easy!
    return np.meshgrid(*list(array_1d for i in range(dim)), indexing=indexing)


def normgrid(arr_1d, dim=2, indexing='xy', ord=None):
    """
    2/23/22 VERIFIED

    Performs a meshgrid over the arr_1d according to ez.meshgrid, then computes the norm = sqrt(x1**2 + x2**2 + ...).

    :param arr_1d: ndarray or list
        The input array/list over which to perform a meshgrid
    :param dim: int
        The number of dimensions
    :param indexing: default 'xy' or 'ij'
        The indexing. Cartesian is default and preferred
    :param ord: None or str
        The type of norm, default is None which gives 2-norm, specify according to np otherwise
    :return: output: ndarray
        The result of the p-norm applied to a meshgrid with p specified by the ord parameter

    --------
    Examples
    --------
    normy = normgrid(ez.linspace(-10, 10, 31), 3) <--- A 3d normgrid
    norm_macdonald = normgrid(ez.linspace(-10, 10, 30), 2) <--- A 2d normgrid

    "I think there's nothing cooler than being a lone wolf. Except at wolf-picnics when you don't have a partner for the
    wolf-wheelbarrow races."
    """
    return np.linalg.norm(meshgrid(arr_1d, dim, indexing=indexing), axis=0, ord=ord)


def shape_mask(arr, r=1, dim=2, shape='circle'):
    """
    2/23/2022 VERIFIED

    A function that produces a circle/square mask based off of an array. The most meaningful way to use this function is
    likely with arr = ez.linspace(-a, a, N).

    :param arr: ndarray (1d array)
        The input array which is is compared to the radius value, just 1d
    :param r: float
        The radius value to compare against (radius of a square = half the width or height in this case)
    :param dim: int
        The dimensionality of the shape mask
    :param shape: {'circle', 'square'}
        The type of shape -- for now just square or circle.
    :return: output: ndarray
        The output mask, a ndarray of bool values
    """
    if shape=='circle':
        norm_arr = normgrid(arr, dim=dim, ord=None)
        return norm_arr <= r
    elif shape=='square':
        norm_arr = normgrid(arr, dim=dim, ord=np.inf)
        return np.abs(norm_arr) <= r


def mesh_spread_2d(x):
    """
    A simple function that just gives the range of a 2d array

    :param x: ndarray
    :return: the range of the array
    """
    return np.amax([abs(x[0,:] - x[-1,:]), abs(x[:,0] - x[:,-1])])


from cvpr_2022_itt_pkg.core import core
from cvpr_2022_itt_pkg.core.core import *
from cvpr_2022_itt_pkg.ece import ece
from cvpr_2022_itt_pkg.ece.ece import *
from cvpr_2022_itt_pkg.physics import physics
from cvpr_2022_itt_pkg.prob import prob


from cvpr_2022_itt_pkg.core.core import *
from cvpr_2022_itt_pkg.ece.ece import *
import numpy as np
import math
import matplotlib.pyplot as plt


# Optics Functions
def fraunhofer(wave, T, D=np.inf, wvl=500e-9, z=10000, mask='circle', pad=0):
    """
    Fraunhofer diffraction. IMPORTANT NOTE: THIS WILL NOT GIVE YOU THE PSF
    FORMATION! IT WILL GIVE YOU FRAUNHOFER DIFFRACTION WITHOUT A LENS.

    For PSF formation with a lens, you have to use one of the ___2psf()
    functions.

    :param wave: ndarray
        The wave over the grid of lengths T x T.
    :param T: float
        The length of one dimension of the grid (assumed to be square grid)
    :param D: float
        Diameter of the aperture. By default set to D=np.inf for plane wave (though, you will get aliasing)
    :param wvl: float
        Wavelength of light (for a general incoherent visible light simulation, choose 555e-9 or something in this
        range)
    :param z: float
        Propagation distance
    :param mask: {'circle', 'square'}
        The mask shape.
    :return: the sample spacing grid; the output wave
    """
    aperture = shape_mask(linspace(-T/2, T/2, wave.shape[0]), r=D/2, dim=2, shape=mask)
    if len(wave.shape) == 3:
        xx, yy, wave_out = fft2(wave * aperture[:,:,np.newaxis], T, pad=pad)
    else:
        xx, yy, wave_out = fft2(wave * aperture, T, pad=pad)
    xx *= (wvl * z)
    yy *= (wvl * z)

    return xx, yy, wave_out


def fresnel(wave, T, D=np.inf, wvl=500e-9, z=10000, mask='circle'):
    """
    Fresnel diffraction. IMPORTANT NOTE: THIS WILL NOT GIVE YOU THE PSF
    FORMATION! IT WILL GIVE YOU Fresnel DIFFRACTION WITHOUT A LENS.

    For PSF formation with a lens, you have to use one of the ___2psf()
    functions.

    :param wave: ndarray
        The wave over the grid of lengths T x T.
    :param T: float
        The length of one dimension of the grid (assumed to be square grid)
    :param D: float
        Diameter of the aperture
    :param wvl: float
        Wavelength of light (for a general incoherent visible light simulation, choose 555e-9 or something in this
        range)
    :param z: float
        Propagation distance
    :param mask: {'circle', 'square'}
        The mask shape. Later implementations will likely remove this or have it as optional.
    :return: the sample spacing grid; the output wave
    """
    if mask=='circle':
        aperture = shape_mask(linspace(-T/2, T/2, wave.shape[0]), r=D/2, dim=2)
    else:
        aperture = shape_mask(linspace(-T/2, T/2, wave.shape[0]), r=D/2, dim=2, shape=mask)
    xprime = normgrid(np.linspace(-T/2, T/2, wave.shape[0]))
    xx, yy, wave_out = fft2(wave * aperture * np.exp(1j*np.pi/(wvl * z) * xprime**2), T)
    xx *= (wvl * z)
    yy *= (wvl * z)

    return xx, yy, wave_out * np.exp(1j * 2*np.pi/wvl * z) / \
           (1j * wvl * z) * np.exp(1j*np.pi * (xx**2 + yy**2) / (wvl * z))


def sph_wave(N, T, L, wvl=500e-9):
    """
    Returns a spherical wave of size NxN over a TxT (m^2) spatial area. There is no amplitude, but you can easily
    multiply the return value

    :param N: int
        Size of the input array in elements (will give output NxN)
    :param T: float
        Metric length of a single dimension (assumes TxT area)
    :param L: float
        Distance of plane at which spherical wave is returned
    :param wvl: float
        The wavelength
    :return: ndarray
        The spherical wave
    """
    xx, yy = meshgrid(linspace(-T/2, T/2, N), dim=2)
    k = 2 * np.pi / wvl
    return np.exp(1j * k * np.sqrt(xx**2 + yy**2 + L**2)) / np.sqrt(xx**2 + yy**2 + L**2)



def phase2psf(phase, T, D, f, wvl=500e-9, L=-1, mask='circle'):
    """
    PSF formation using the phase of the wave directly. Calls upon Fraunhofer and is evaluated at the focal distance,
    with the option of the distance L being provided to the function to know the PSF size in terms of the dimensions
    of the source object. Otherwise, dimensions returned will be in focal plane.

    :param phase: ndarray
        The phase of the wave. Often in this simulation package, only the phase matters, so this is a
        straightforward way of going from phase to space
    :param T: float
        Length of the input grid in meters
    :param D: float
        Diameter of the aperture in meters
    :param f: float
        Focal length in meters
    :param wvl: float
        Wavelength of the incident wave in meters
    :param L: float
        Propagation distance in meters.
    :return: spatial grid; psf = abs(wave_out)**2
    """
    xx, yy, wave_out = fraunhofer(np.exp(1j * phase), T=T, D=D, wvl=wvl, z=f, mask=mask)
    psf = np.abs(wave_out)**2
    if L == -1:
        return xx, yy, psf
    else:
        return xx*(L/f), yy*(L/f), psf


def wave2psf(wave, T, D, f, wvl=500e-9, L=-1, mask='circle'):
    """
    PSF formation using the wave. Calls upon Fraunhofer and is evaluated at the focal distance, with the option of
    the distance L being provided to the function to know the PSF size in terms of the dimensions of the source object.

    :param wave: ndarray
        The phase of the wave. Often in this simulation package, only the phase matters, so this is a
        straightforward way of going from phase to space
    :param T: float
        Length of the input grid in meters
    :param D: float
        Diameter of the aperture in meters
    :param f: float
        Focal length in meters
    :param wvl: float
        Wavelength of the incident wave in meters
    :param L: float
        Propagation distance in meters
    :return: spatial grid; psf = abs(wave_out)**2
    """
    xx, yy, wave_out = fraunhofer(wave, T, D, wvl=wvl, z=f, mask=mask)
    psf = np.abs(wave_out)**2
    if L == -1:
        return xx, yy, psf
    else:
        return xx * (L / f), yy * (L / f), psf


def otf_calc(pup_fun, T, D, f, wvl):
    '''
    A function that accepts the generalized pupil function and returns the otf with frequency components

    :param pup_fun: ndarray
        The generalized pupil array. Use in the form aperture_shape * np.exp(1j * phase_distortion)
    :param T: float
        The measure of the window of the pupil function [meters]
    :param D: float
        The size of the aperture [meters]
    :param f: float
        The focal length [meters]
    :param wvl: float
        The wavelength [meters]
    :return: (ndarray, ndarray, ndarray)
        The normalized OTF with frequency meshgrid
    '''
    w = D/2
    f_cutoff = w / (wvl * f)
    a, a, P = fft2(pup_fun, 1)
    a, a, otf = np.abs(ifft2(P * P, 1))
    xx, yy = meshgrid(linspace(-1/2, 1/2, pup_fun.shape[0]))
    scaling = 2*D/T * f_cutoff
    return xx/scaling, yy/scaling, otf/np.amax(otf)


def zern_phase_rep(phase, num_zern=36):
    print('hi')


def gen_rand_vec(R, n_samp, mu=0, mode='chol'):
    """
    Generating a random vector according to an covariance matrix.

    :param R: ndarray
        Covariance matrix
    :param n_samp: int
        Number of samples desired
    :param mu: float
        Mean value of the random vector
    :param mode: {'chol', 'eig'}
        Different generation approaches, Cholesky by default
    :return: the n_samp desired random vectors
    """

    if mode == 'chol':
        L = np.linalg.cholesky(R)
        return L @ np.random.randn(R.shape[1], n_samp) + mu
    if mode == 'eig':
        vals, vecs = np.linalg.eig(R)
        print(vecs)
        return (vecs @ np.diag(np.sqrt(vals))) @ np.random.randn(R.shape[1], n_samp) + mu


def defocus_abgen(norm_grid, L_true, L_selected, wvl=500e-9):
    '''
    A function that returns a defocus (difference of spheres) aberration in radians.

    :param xx: ndarray
        input norm_grid (this function will square the input -- the argument should measure distance i.e. sqrt(
        something)
    :param L_true: float
        the true distance of the point source to the camera
    :param L_selected: float
        the selected distance of the point source (the "wrong" focus distance)
    :param wvl: float
        wavelength of the light
    :return: ndarray
        the output defocus aberration in radians
    '''
    return -1/2 * (1/L_selected - 1/L_true) * norm_grid**2 * (2*np.pi/wvl)

def genZernikeCoeff(num_zern, D_r0, num_vecs = 1):
    '''
    Just a simple function to generate random coefficients as needed, conforms to Noll's Theory. The nollCovMat()
    function is at the heart of this function.

    A note about the function call of nollCovMat in this function. The input (..., 1, 1) is done for the sake of
    flexibility. One can call the function in the typical way as is stated in its description. However, for
    generality, the D/r0 weighting is pushed to the "b" random vector, as the covariance matrix is merely scaled by
    such value.

    :param num_zern: This is the number of Zernike basis functions/coefficients used. Should be numbers that the pyramid
    rows end at. For example [1, 3, 6, 10, 15, 21, 28, 36]
    :param D_r0:
    :return:
    '''
    C = nollCovMat(num_zern)
    a = gen_rand_vec(C, num_vecs) * D_r0 ** (5.0/6.0)

    return a


def nollCovMat(Z):
    """
    This function generates the covariance matrix for a single point source. See the associated paper for details on
    the matrix itself.

    :param Z: int
        Number of Zernike basis functions/coefficients, determines the size of the matrix.
    :param D: float
        The diameter of the aperture (meters)
    :param fried: float
        The Fried parameter value
    :return: the Noll covariance matrix
    """
    C = np.zeros((Z,Z))
    for i in range(Z):
        for j in range(Z):
            ni, mi = nollToZernInd(i+1)
            nj, mj = nollToZernInd(j+1)
            if (abs(mi) == abs(mj)) and (np.mod(i - j, 2) == 0):
                num = math.gamma(14.0/3.0) * math.gamma((ni + nj - 5.0/3.0)/2.0)
                den = math.gamma((-ni + nj + 17.0/3.0)/2.0) * math.gamma((ni - nj + 17.0/3.0)/2.0) * \
                      math.gamma((ni + nj + 23.0/3.0)/2.0)
                coef1 = 0.0072 * (np.pi ** (8.0/3.0)) * np.sqrt((ni + 1) * (nj + 1)) * \
                        ((-1) ** ((ni + nj - 2*abs(mi))/2.0))
                C[i, j] = coef1*num/den
            else:
                C[i, j] = 0
    C[0,0] = 1
    return C


def zernikeGen(N, coeff, axis=-1):
    """
    NEED TO UPDATE

    Generating the Zernike phase representation from specified coefficients.

    :param N: int
        1d length of array, the phase will be NxN
    :param coeff: ndarray
        The coefficient array, there can be multiple vectors, resulting in the output being 4d
    :param axis: int
        The axis over which the coefficient array is used to compute the phase components
    :return: the phase realization
    """

    num_coeff = coeff.shape[axis]

    # Setting up 2D grid
    x_grid, y_grid = meshgrid(np.linspace(-1, 1, N))

    zern_out = np.zeros((N,N,num_coeff))
    for i in range(num_coeff):
        zern_out[:,:,i] = coeff[i]*genZernPoly(i+1, x_grid, y_grid)

    return zern_out


def nollToZernInd(j):
    """
    This function maps the input "j" to the (row, column) of the Zernike pyramid using the Noll numbering scheme.

    Authors: Tim van Werkhoven, Jason Saredy
    See: https://github.com/tvwerkhoven/libtim-py/blob/master/libtim/zern.py
    """
    if (j == 0):
        raise ValueError("Noll indices start at 1, 0 is invalid.")
    n = 0
    j1 = j-1
    while (j1 > n):
        n += 1
        j1 -= n
    m = (-1)**j * ((n % 2) + 2 * int((j1+((n+1)%2)) / 2.0 ))

    return n, m


def genZernPoly(index, x_grid, y_grid):
    """
    This function generates overall Zernike polynomials according to Noll's definition, which is a modified Zernike
    polynomial.

    :param index: int
        The Zernike index under consideration
    :param x_grid: ndarray
        The x_grid over which the Zernike polynomials are computed
    :param y_grid: ndarray
        The y_grid over which the Zernike polynomials are computed
    :return: the Zernike polynomial
    """
    n,m = nollToZernInd(index)
    radial = radialZernike(x_grid, y_grid, (n,m))
    if m == 0:
        return np.sqrt(n+1)*radial
    if np.mod(index,2) == 0:
        return np.multiply(np.sqrt(n+1)*radial, np.sqrt(2)*np.cos(m * np.arctan2(y_grid, x_grid)))
    if np.mod(index,2) == 1:
        return np.multiply(np.sqrt(n+1)*radial, np.sqrt(2)*np.sin(m * np.arctan2(y_grid, x_grid)))
    '''
    if m < 0:
        return np.multiply(np.sqrt(n+1)*radial, np.sqrt(2)*np.sin(-m * np.arctan2(y_grid, x_grid)))
    else:
        return np.multiply(np.sqrt(n+1)*radial, np.sqrt(2)*np.cos(m * np.arctan2(y_grid, x_grid)))
    '''

def radialZernike(x_grid, y_grid, z_ind):
    """
    Generating the individual radial polynomial component of the Zernike polynomial.

    :param x_grid: ndarray
        The x_grid over which the Zernike polynomials are computed
    :param y_grid: ndarray
        The y_grid over which the Zernike polynomials are computed
    :param z_ind: tuple
        The (radial,angular) pair that parameterizes the Zernike representation.
    :return:
    """

    rho = np.sqrt(x_grid ** 2 + y_grid ** 2)
    radial = np.zeros(rho.shape)
    n = z_ind[0]
    m = np.abs(z_ind[1])

    for k in range(int((n - m)/2 + 1)):
        temp = (-1) ** k * math.factorial(n - k) / (math.factorial(k) * math.factorial((n + m)/2 - k)
                                                       * math.factorial((n - m)/2 - k))
        radial += temp * rho ** (n - 2*k)
    return radial


import numpy as np

# Random Functions
def pdf(x):
    """
    Generating the probability density function (pdf) with auto bin widths. This function integrates to 1, but does not
    SUM to one, with the exception of bins are of width 1.

    :param x: ndarray
        Random Samples to plot pdf over
    :return: pdf values; center of bin locations; bin widths
    """

    vals, bins = np.histogram(x, bins='auto', density=True)
    midbins = (bins[1:] + bins[:-1]) / 2
    return vals, midbins, np.abs(midbins[0] - midbins[1])


def cdf(x):
    """
        Generating the probability density function (pdf) with auto bin widths. This function integrates to 1, but does not
        SUM to one, with the exception of bins are of width 1.

        :param x: ndarray
            Random Samples to plot pdf over
        :return: cdf values; center of bin locations
        """
    vals, midbins, dx = pdf(x)
    return np.cumsum(vals)*dx, midbins


def gen_dist(x, cdf, n_samp):
    """
    Generating a distribution from its cdf. The cdf in this case can be empirical or theoretical.

    :param x: ndarray
        This should be the x-locations of the bins, i.e. cdf = f(x).
    :param cdf:
        The cdf values itself. They should go from 0 to 1, there is an additional check for that within this function.
    :param n_samp:
        The number of samples desired to be generated.
    :return: the n_samp random samples
    """
    cdf[0] = 0
    cdf[-1] = 1
    seed = np.random.rand(n_samp)
    vals = np.interp(seed, cdf, x) #Intentionally swapped from how you'd think! (Inverse CDF rule)
    return vals


def gen_rand_vec(R, n_samp, mu=0, mode='chol'):
    """
    Generating a random vector according to an covariance matrix.

    :param R: ndarray
        Covariance matrix
    :param n_samp: int
        Number of samples desired
    :param mu: float
        Mean value of the random vector
    :param mode: {'chol', 'eig'}
        Different generation approaches, Cholesky by default
    :return: the n_samp desired random vectors
    """

    if mode == 'chol':
        L = np.linalg.cholesky(R)
        return L @ np.random.randn(R.shape[1], n_samp) + mu
    if mode == 'eig':
        vals, vecs = np.linalg.eig(R)
        print(vecs)
        return (vecs @ np.diag(np.sqrt(vals))) @ np.random.randn(R.shape[1], n_samp) + mu


def emp_cov(vecs, force_symm=False):
    """
    Computing the empirical covariance matrix. Can force the result to by positive semidefinite if desired.

    :param vecs:
    :param force_symm: bool
        Not done yet
    :return:
    """

    emp_mu = np.mean(vecs, axis=-1)
    emp_mu = np.expand_dims(emp_mu, axis=-1)
    return (vecs - emp_mu) @ (vecs - emp_mu).T / vecs.shape[-1]


def gen_rand_2dfield_psd(psd, T=1, N=1):
    """
    Generating N 2d random fields from its PSD. The default value of T = 1, representing the distance in the Fourier
    space, however, can be provided different value giving the resultant spatial sampling in the outputs.

    :param psd: ndarray
        Power spectral density
    :param T: float
        1d length in the Fourier space that the PSD is generated over
    :param N: int
        Number of output random fields
    :return: the output random fields and spatial grids
    """
    a,a,temp = fft2(np.random.randn(*(psd.shape + (N,))) + 1j * np.random.randn(*(psd.shape + (N,))), T, norm='dft')
    return ifft2(temp * np.sqrt(psd.shape[0]*psd.shape[1]) * np.sqrt(psd[:,:,np.newaxis]), T, norm='dft')


def gen_rand_2dfield_cov(cov, T=1, N=1):
    """
    Generating N 2d random fields from its covariance matrix. The default value of T = 1, representing the distance in
    the spatial domain, however, can be provided different value giving the resultant spatial sampling in the
    outputs.

    :param psd: ndarray
        Power spectral density
    :param T: float
        1d length in the Fourier space that the PSD is generated over
    :param N: int
        Number of output random fields
    :return: the output random fields and spatial grids
    """
    x, y, psd = fft2(cov, T, norm='dft')
    x, y, fields = gen_rand_2dfield_psd(np.abs(psd), mesh_spread_2d(x), N)
    return x, y, fields/np.sqrt(2)



from cvpr_2022_itt_pkg.core.core import *
import numpy as np


'''
This is basically a laundry list of every ECE thing I want to do but have to use multiple commands for/Google how to 
do it and it is really annoying so I have this instead. Uses the core module.
'''


# ECE Functions
def fft(y, T, pad=0, axis=- 1, norm='dft'):
    """
    A 2-sided (positive and negative) fft that uses numpy's fft as a starting point. Contains normalizations to
    match the dft and cft for assistance in visualization. It is often the case in optical simulations that a 1-way
    fourier transforms is used to compute further *spatial* behavior, so both the sample spacing and amplitude must
    be handled appropriately and not thrown away.

    :param y: ndarray
        Input signal to be fft'd
    :param T: float
        Length of interval in space/time/etc.
    :param pad: int
        Size of TOTAL pad. This means a signal [1, 1] with pad = 2 with be padded to be [0, 1, 1, 0]. This should
        match any ifft you did before!!
    :param axis: tuple
        Specifies which axis should the fft be taken over
    :param norm: {'dft', 'cft'}
        Specifies how the fft should be normalized, to match the dft or cft. See examples for some fine details
    :return: xf: fourier grid; yf: fft signal
    """
    # Just trying to make sure it won't mess up
    N = np.squeeze(np.array(y.shape))

    # Padding the signal according to the specified pad. Note that the pad passed to the function will be divided by
    # 2, and padded to the signal. A signal [1, 1] with pad = 2 with be padded to be [0, 1, 1, 0]
    y = np.pad(y, pad//2, mode='constant')
    xf = np.linspace(-0.5 * N / T, 0.5 * N / T, N+pad)
    yf = 1 / N * np.fft.fftshift(np.fft.fft(np.fft.ifftshift(y, axis=axis), axis=axis), axis=axis)

    # Returning the proper output according to the normalization type specified
    if norm == 'dft':
        return xf[:N+pad], yf[:N+pad]
    if norm == 'cft':
        return xf[:N + pad], T*yf[:N + pad]


def ifft(y, T, pad=0, axis=- 1, norm='dft'):
    """
    A 2-sided (positive and negative) ifft that uses numpy's fft as a starting point. Contains
    normalizations to match the dft and cft for assistance in visualization. It is often the case in optical simulations
    that a 1-way fourier transforms is used to compute further *spatial* behavior, so both the sample spacing and
    amplitude must be handled appropriately and not thrown away.

    :param y: ndarray
        Input signal to be ifft'd
    :param T: float
        Length of interval in fourier space (2 x max_frequency)
    :param pad: int
        Tells the function how to interpret the signal for sample spacing. This should match any fft you did before!!
    :param axis: tuple
        Specifies which axis should the fft be taken over
    :param norm: {'dft', 'cft'}
        Specifies how the fft should be normalized, to match the dft or cft. See examples for some fine details
    :return: xf: inverse-fourier grid; yf: ifft signal
    """
    # Just trying to make sure it won't mess up
    N = np.squeeze(np.array(y.shape))
    ogN = N - pad # The original N value

    xf = np.linspace(-0.5 * N / T, 0.5 * N / T, ogN)
    yf = ogN * np.fft.fftshift(np.fft.ifft(np.fft.ifftshift(y, axis=axis), axis=axis), axis=axis)

    # Returning the proper output according to the normalization type specified
    if norm == 'dft':
        return xf, yf[pad//2:N-pad//2]
    if norm == 'cft':
        return xf, yf[pad//2:N-pad//2]/T


def fft2(y, T, pad=0, axes=(0,1), norm='dft'):
    """
    A function that uses numpy's fft2 as a starting point. Contains normalizations to match the dft and cft for
    assistance in visualization/computation. It is often the case in optical simulations
    that a 1-way fourier transforms is used to compute further *spatial* behavior, so both the sample spacing and
    amplitude must be handled appropriately and not thrown away.

    :param y: ndarray
        Input signal to be fft'd
    :param T: float
        Length of interval in space/time/etc.
    :param pad: int
        Size of TOTAL pad. This means a signal [1, 1] with pad = 2 with be padded to be [0, 1, 1, 0]. This should
        match any ifft you did before!!
    :param axis: tuple
        Specifies which axis should the fft be taken over
    :param norm: {'dft', 'cft'}
        Specifies how the fft should be normalized, to match the dft or cft. See examples for some fine details
    :return: xf: fourier grid; yf: fft signal
    """
    # Just trying to make sure it won't mess up
    N = np.squeeze(np.array(y.shape))

    # Padding the signal according to the specified pad. Note that the pad passed to the function will be divided by
    # 2, and padded to the signal. A signal [1, 1] with pad = 2 with be padded to be [0, 1, 1, 0]
    y = np.pad(y, (pad//2,pad//2), mode='constant')
    if N[0] == N[1]:
        xf = np.linspace(-0.5 * N[0] / T, 0.5 * N[0] / T, N[0]+pad)
        xxf, yyf = np.meshgrid(xf[:N[0]+pad], xf[:N[0]+pad])
        yf = 1.0 / (N[0]*N[1]) * np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(y, axes=axes), axes=axes), axes=axes)

        # Returning the proper output according to the normalization type specified
        if norm == 'dft':
            return xxf, yyf, yf[:N[0]+pad, :N[1]+pad]
        if norm == 'cft':
            return xxf, yyf, yf[:N[0]+pad, :N[1]+pad] * T**2


def ifft2(y, T, pad=0, axes=(0,1), norm='dft'):
    """
    A function that uses numpy's ifft2 as a starting point. Contains normalizations to match the dft and cft for
    assistance in visualization. It is often the case in optical simulations that a 1-way fourier transforms is used
    to compute further *spatial* behavior, so both the sample spacing and amplitude must be handled appropriately
    and not thrown away.

    :param y: ndarray
        Input signal to be ifft'd
    :param T: float
        Length of interval in space/time/etc.
    :param pad: int
        Tells the function how to interpret the signal for sample spacing. This should match any fft you did before!!
    :param axis: tuple
        Specifies which axis should the fft be taken over
    :param norm: {'dft', 'cft'}
        Specifies how the fft should be normalized, to match the dft or cft. See examples for some fine details
    :return: xf, yyf: inverse-fourier grids; yf: ifft signal
    """
    # Just trying to make sure it won't mess up
    N = np.squeeze(np.array(y.shape))
    ogN = N - pad # The original N value
    if N[0] == N[1]:
        xf = np.linspace(-0.5 * N[0] / T, 0.5 * N[0] / T, ogN[0])
        xxf, yyf = np.meshgrid(xf, xf)
        yf = (ogN[0] * ogN[0]) * np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(y, axes=axes), axes=axes), axes=axes)

        # Returning the proper output according to the normalization type specified
        if norm == 'dft':
            return xxf, yyf, yf[pad//2:N[0]-pad//2, pad//2:N[1]-pad//2]
        if norm == 'cft':
            return xxf, yyf, yf[pad//2:N[0]-pad//2, pad//2:N[1]-pad//2] / T**2


def basis_decomp(sig, dA, bases):
    if len(bases.shape) == 2:
        print('TODO')
    if len(bases.shape) == 3:
        if sig.shape == bases[:2]:
            return np.einsum('ij,ijk->k', sig, bases) * dA
        if sig.shape == bases[-2:]:
            return np.einsum('ij,kji->k', sig, bases) * dA


